# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import OrderedDict

from .base import BaseProvider


class HelmProvider(BaseProvider):
    tag = "helm"

    def __init__(self, body, authenticators, **kwargs):
        super(HelmProvider, self).__init__(body)
        self.tag = kwargs.get("tag", self.tag)
        self.vars = body.get("vars", {})
        self.version = body.get("version")

    def hcl(self):
        result = []
        provider_vars = {}
        try:
            for k, v in self.vars.items():
                provider_vars[k] = v
        except (KeyError, TypeError):
            """No provider vars were set."""
            pass

        result.append(f'provider "{self.tag}" {{')
        for k, v in provider_vars.items():

            # Handle special case for kubernetes block in helm provider
            if k.lower() == "kubernetes" and isinstance(v, OrderedDict):
                result.append(f"  {k} {{")
                for ik, iv in v.items():
                    if iv and '"' not in iv:
                        result.append(f'    {ik} = "{iv}"')
                    else:
                        result.append(f"    {ik} = {iv}")
                result.append("  }")
                continue

            if v and '"' not in v:
                result.append(f'  {k} = "{v}"')
            else:
                result.append(f"  {k} = {v}")
        result.append("}")

        return "\n".join(result)



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import glob
import os
import shutil
import urllib
import zipfile

import click
from tenacity import retry, stop_after_attempt, wait_chain, wait_fixed
from tfworker.commands.root import get_platform


class PluginsCollection(collections.abc.Mapping):
    def __init__(self, body, temp_dir):
        self._plugins = body
        self._temp_dir = temp_dir

    def __len__(self):
        return len(self._providers)

    def __getitem__(self, value):
        if type(value) == int:
            return self._providers[list(self._providers.keys())[value]]
        return self._providers[value]

    def __iter__(self):
        return iter(self._providers.values())

    @retry(
        wait=wait_chain(
            wait_fixed(5),
            wait_fixed(30),
            wait_fixed(60),
            wait_fixed(180),
            wait_fixed(300),
        ),
        stop=stop_after_attempt(5),
    )
    def download(self):
        """
        Download the required plugins.

        This could be further optimized to not download plugins from hashicorp,
        but rather have them in a local repository or host them in s3, and get
        them from an internal s3 endpoint so no transit charges are incurred.
        Ideally these would be stored between runs, and only downloaded if the
        versions have changed. In production try  to remove all all external
        repositories/sources from the critical path.
        """
        opsys, machine = get_platform()
        _platform = f"{opsys}_{machine}"

        plugin_dir = f"{self._temp_dir}/terraform-plugins"
        if not os.path.isdir(plugin_dir):
            os.mkdir(plugin_dir)

        for name, details in self._plugins.items():
            uri = get_url(name, details)
            file_name = uri.split("/")[-1]

            click.secho(
                f"getting plugin: {name} version {details['version']} from {uri}",
                fg="yellow",
            )

            with urllib.request.urlopen(uri) as response, open(
                f"{plugin_dir}/{file_name}", "wb"
            ) as plug_file:
                shutil.copyfileobj(response, plug_file)
            with zipfile.ZipFile(f"{plugin_dir}/{file_name}") as zip_file:
                zip_file.extractall(f"{plugin_dir}/{_platform}")
            os.remove(f"{plugin_dir}/{file_name}")

            files = glob.glob(f"{plugin_dir}/{_platform}/terraform-provider*")
            for afile in files:
                os.chmod(afile, 0o755)
                filename = os.path.basename(afile)
                os.rename(afile, f"{plugin_dir}/{filename}")

            click.secho(f"plugin installed to: {plugin_dir}/{_platform}/", fg="yellow")


def get_url(name, details):
    """
    Determine the URL for the plugin

    get URL returns a fully qualifed URL, including the file name.

    In order to support third party terraform plugins we can not
    assume the hashicorp repository. It will function as a default,
    but if baseURL is provided in the plugin settings it will be
    used instead. The logic to determine the complete remote path
    will also be here to simplify the logic in the download method.
    """
    opsys, machine = get_platform()
    _platform = f"{opsys}_{machine}"

    try:
        version = details["version"]
    except KeyError:
        raise KeyError(f"version must be specified for plugin {name}")

    # set the file name, allow it to be overridden with key "filename"
    default_file_name = f"terraform-provider-{name}_{version}_{_platform}.zip"
    file_name = details.get("filename", default_file_name)

    # set the base url, allow it to be overridden with key "baseURL"
    default_base_url = (
        f"https://releases.hashicorp.com/terraform-provider-{name}/{version}"
    )
    base_uri = details.get("baseURL", default_base_url).rstrip("/")

    return f"{base_uri}/{file_name}"



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tfworker.commands.base import BaseCommand


class CleanCommand(BaseCommand):
    def __init__(self, rootc, *args, **kwargs):
        self._config = rootc.config
        self._deployment = kwargs.get("deployment")
        self._limit = kwargs.get("limit", [])
        super(CleanCommand, self).__init__(rootc, **kwargs)

    def exec(self):
        for prov in self._providers:
            prov.clean(self._deployment, self._limit)



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import Backends
from .gcs import GCSBackend  # noqa
from .s3 import S3Backend  # noqa


def select_backend(backend, deployment, authenticators, definitions):
    if backend == Backends.s3:
        return S3Backend(authenticators, definitions, deployment=deployment)
    elif backend == Backends.gcs:
        return GCSBackend(authenticators, definitions, deployment=deployment)



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import copy

from .aws import AWSProvider  # noqa
from .base import UnknownProvider  # noqa
from .generic import GenericProvider  # noqa
from .google import GoogleProvider  # noqa
from .google_beta import GoogleBetaProvider  # noqa
from .helm import HelmProvider  # noqa

ALL = [AWSProvider, GoogleProvider, GoogleBetaProvider, HelmProvider]


class ProvidersCollection(collections.abc.Mapping):
    def __init__(self, providers_odict, rootc):
        provider_map = dict([(prov.tag, prov) for prov in ALL])
        self._providers = copy.deepcopy(providers_odict)
        for k, v in self._providers.items():
            try:
                self._providers[k] = provider_map[k](v, rootc)

            except KeyError:
                self._providers[k] = GenericProvider(v, tag=k)

    def __len__(self):
        return len(self._providers)

    def __getitem__(self, value):
        if type(value) == int:
            return self._providers[list(self._providers.keys())[value]]
        return self._providers[value]

    def __iter__(self):
        return iter(self._providers.values())

    def hcl(self):
        return "\n".join([prov.hcl() for _, prov in self._providers.items()])



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

from tfworker.authenticators import AuthenticatorsCollection
from tfworker.backends import select_backend
from tfworker.definitions import DefinitionsCollection
from tfworker.plugins import PluginsCollection
from tfworker.providers import ProvidersCollection


class BaseCommand:
    def __init__(
        self, rootc, deployment="undefined", limit=None, plan_for="apply", **kwargs
    ):
        self._version = None
        self._providers = None
        self._definitions = None
        self._backend = None
        self._plugins = None
        self._template_vars = OrderedDict()
        self._terraform_vars = OrderedDict()
        self._remote_vars = OrderedDict()

        self._temp_dir = rootc.temp_dir
        self._repository_path = rootc.args.repository_path
        self._authenticators = AuthenticatorsCollection(
            rootc.args, deployment=deployment, **kwargs
        )

        rootc.clean = kwargs.get("clean", True)

        self._providers = ProvidersCollection(
            rootc.providers_odict, self._authenticators
        )
        self._definitions = DefinitionsCollection(
            rootc.definitions_odict,
            deployment,
            limit,
            plan_for,
            self._providers,
            self._repository_path,
            rootc,
            self._temp_dir,
        )
        self._plugins = PluginsCollection(rootc.plugins_odict, self._temp_dir)
        self._backend = select_backend(
            rootc.args.backend,
            deployment,
            self._authenticators,
            self._definitions,
        )

    @property
    def authenticators(self):
        return self._authenticators

    @property
    def backend(self):
        return self._backend

    @property
    def providers(self):
        return self._providers

    @property
    def definitions(self):
        return self._definitions

    @property
    def plugins(self):
        return self._plugins

    @property
    def temp_dir(self):
        return self._temp_dir

    @property
    def repository_path(self):
        return self._repository_path



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shlex

import boto3
from tfworker import constants as const

from .base import BaseAuthenticator


class AWSAuthenticator(BaseAuthenticator):
    tag = "aws"

    def __init__(self, state_args, **kwargs):
        super(AWSAuthenticator, self).__init__(state_args, **kwargs)

        self.access_key_id = self._resolve_arg("aws_access_key_id")
        self.backend_region = self._resolve_arg("backend_region")
        self.bucket = self._resolve_arg("backend_bucket")
        self.prefix = self._resolve_arg("backend_prefix")
        self.profile = self._resolve_arg("aws_profile")
        self.region = self._resolve_arg("aws_region")
        self.role_arn = self._resolve_arg("role_arn")
        self.secret_access_key = self._resolve_arg("aws_secret_access_key")
        self.session_token = self._resolve_arg("aws_session_token")

        self.deployment = kwargs.get("deployment")

        self._account_id = None
        self._backend_session = None
        self._session = None

        # If the default value is used, render the deployment name into it
        if self.prefix == const.DEFAULT_BACKEND_PREFIX:
            self.prefix = const.DEFAULT_BACKEND_PREFIX.format(
                deployment=self.deployment
            )

        aws_is_active = (self.access_key_id and self.secret_access_key) or self.profile
        if aws_is_active:
            # Initialize the session objects
            self._session = boto3.Session(
                region_name=self.region, **self._session_state_args
            )

            if not self.role_arn:
                # if a role was not provided, need to ensure credentials areset
                # in the config, these will come from the session
                self.access_key_id = self._session.get_credentials().access_key
                self.secret_access_key = self._session.get_credentials().secret_key
                self.session_token = self._session.get_credentials().token

                if self.backend_region == self.region:
                    self._backend_session = self._session
                else:
                    self._backend_session = boto3.Session(
                        region_name=self.backend_region, **self._session_state_args
                    )
            else:
                (self.__session, creds) = AWSAuthenticator.get_assumed_role_session(
                    self._session, self.self.role_arn
                )
                self.access_key_id = creds["AccessKeyId"]
                self.secret_access_key = creds["SecretAccessKey"]
                self.session_token = creds["SessionToken"]

                if self.backend_region == self.region:
                    self._backend_session = self._session
                else:
                    self._backend_session = boto3.Session(
                        region_name=self.backend_region,
                        aws_access_key_id=self.access_key_id,
                        aws_secret_access_key=self.secret_access_key,
                        aws_session_token=self.session_token,
                    )

    @property
    def _session_state_args(self):
        state_args = dict()

        if self.profile:
            state_args["profile_name"] = self.profile

        if self.access_key_id:
            state_args["aws_access_key_id"] = self.access_key_id

        if self.secret_access_key:
            state_args["aws_secret_access_key"] = self.secret_access_key

        if self.session_token is not None:
            state_args["aws_session_token"] = self.session_token

        return state_args

    @property
    def account_id(self):
        if not self._account_id:
            self._account_id = AWSAuthenticator.get_aws_id(
                self.access_key_id,
                self.secret_access_key,
                self.session_token,
            )
        return self._account_id

    @property
    def backend_session(self):
        return self._backend_session

    @property
    def session(self):
        return self._session

    def env(self):
        result = {}
        if self.access_key_id:
            result["AWS_ACCESS_KEY_ID"] = shlex.quote(self.access_key_id)
        if self.region:
            result["AWS_DEFAULT_REGION"] = shlex.quote(self.region)
        if self.secret_access_key:
            result["AWS_SECRET_ACCESS_KEY"] = shlex.quote(self.secret_access_key)
        if self.session_token:
            result["AWS_SESSION_TOKEN"] = shlex.quote(self.session_token)
        return result

    @staticmethod
    def get_assumed_role_session(
        session, role_arn, session_name="AssumedRoleSession1", duration=3600
    ):
        """ get_assumed_role_session returns a boto3 session updated with assumed role credentials """
        sts_client = session.client("sts")
        role_creds = sts_client.assume_role(
            RoleArn=role_arn, RoleSessionName=session_name, DurationSeconds=duration
        )["Credentials"]

        new_session = boto3.Session(
            aws_access_key_id=role_creds["AccessKeyId"],
            aws_secret_access_key=role_creds["SecretAccessKey"],
            aws_session_token=role_creds["SessionToken"],
        )

        return new_session, role_creds

    @staticmethod
    def get_aws_id(key_id, key_secret, session_token=None):
        """Return the AWS account ID."""
        client = boto3.client(
            "sts",
            aws_access_key_id=key_id,
            aws_secret_access_key=key_secret,
            aws_session_token=session_token,
        )
        return client.get_caller_identity()["Account"]



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import boto3
import click

from .base import BaseBackend


class S3Backend(BaseBackend):
    tag = "s3"
    auth_tag = "aws"

    def __init__(self, authenticators, definitions, deployment=None):
        self._authenticator = authenticators[self.auth_tag]
        self._definitions = definitions
        self._deployment = "undefined"
        if deployment:
            self._deployment = deployment

        # Create locking table for aws backend

        click.secho(
            f"Creating backend locking table: terraform-{deployment}", fg="yellow"
        )
        self.create_table(
            f"terraform-{deployment}",
            self._authenticator.backend_region,
            self._authenticator.access_key_id,
            self._authenticator.secret_access_key,
            self._authenticator.session_token,
        )

    @staticmethod
    def create_table(
        name,
        region,
        key_id,
        key_secret,
        session_token,
        read_capacity=1,
        write_capacity=1,
    ):
        """Create a dynamodb table."""
        client = boto3.client(
            "dynamodb",
            region_name=region,
            aws_access_key_id=key_id,
            aws_secret_access_key=key_secret,
            aws_session_token=session_token,
        )
        tables = client.list_tables()
        table_key = "LockID"
        if name in tables["TableNames"]:
            click.secho("DynamoDB lock table found, continuing.", fg="yellow")
        else:
            click.secho(
                "DynamoDB lock table not found, creating, please wait...", fg="yellow"
            )
            client.create_table(
                TableName=name,
                KeySchema=[{"AttributeName": table_key, "KeyType": "HASH"}],
                AttributeDefinitions=[
                    {"AttributeName": table_key, "AttributeType": "S"},
                ],
                ProvisionedThroughput={
                    "ReadCapacityUnits": read_capacity,
                    "WriteCapacityUnits": write_capacity,
                },
            )

            client.get_waiter("table_exists").wait(
                TableName=name, WaiterConfig={"Delay": 10, "MaxAttempts": 30}
            )

    def hcl(self, name):
        state_config = []
        state_config.append("terraform {")
        state_config.append('  backend "s3" {')
        state_config.append(f'    region = "{self._authenticator.backend_region}"')
        state_config.append(f'    bucket = "{self._authenticator.bucket}"')
        state_config.append(
            f'    key = "{self._authenticator.prefix}/{name}/terraform.tfstate"'
        )
        state_config.append(f'    dynamodb_table = "terraform-{self._deployment}"')
        state_config.append('    encrypt = "true"')
        state_config.append("  }")
        state_config.append("}")
        return "\n".join(state_config)

    def data_hcl(self, exclude):
        remote_data_config = []
        # Call the iter method for explicit control of iteration order
        for definition in self._definitions.iter():
            if definition.tag == exclude:
                break
            remote_data_config.append(
                f'data "terraform_remote_state" "{definition.tag}" {{'
            )
            remote_data_config.append('  backend = "s3"')
            remote_data_config.append("  config = {")
            remote_data_config.append(
                f'    region = "{self._authenticator.backend_region}"'
            )
            remote_data_config.append(f'    bucket = "{self._authenticator.bucket}"')
            remote_data_config.append(
                "    key ="
                f' "{self._authenticator.prefix}/{definition.tag}/terraform.tfstate"'
            )
            remote_data_config.append("  }")
            remote_data_config.append("}\n")
        return "\n".join(remote_data_config)



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shlex

from tfworker import constants as const

from .base import BaseAuthenticator


class GoogleAuthenticator(BaseAuthenticator):
    tag = "google"

    def __init__(self, state_args, **kwargs):
        super(GoogleAuthenticator, self).__init__(state_args, **kwargs)

        self.bucket = self._resolve_arg("backend_bucket")
        self.creds_path = self._resolve_arg("gcp_creds_path")
        self.prefix = self._resolve_arg("backend_prefix")
        self.project = self._resolve_arg("gcp_project")
        self.region = self._resolve_arg("gcp_region")

        self.deployment = kwargs.get("deployment")

        if self.prefix == const.DEFAULT_BACKEND_PREFIX:
            self.prefix = const.DEFAULT_BACKEND_PREFIX.format(
                deployment=self.deployment
            )

    def env(self):
        result = {}
        if self.creds_path:
            result["GOOGLE_APPLICATION_CREDENTIALS"] = shlex.quote(self.creds_path)
        return result



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
import os
import re
import shlex
import shutil
import subprocess

import click
from tfworker.commands.base import BaseCommand


class HookError(Exception):
    pass


class PlanChange(Exception):
    pass


class TerraformError(Exception):
    pass


class TerraformCommand(BaseCommand):
    def __init__(self, rootc, **kwargs):
        self._destroy = kwargs.get("destroy")
        self._tf_apply = kwargs.get("tf_apply")
        if self._tf_apply and self._destroy:
            click.secho("can not apply and destroy at the same time", fg="red")
            raise SystemExit(1)

        self._b64_encode = kwargs.get("b64_encode")
        self._deployment = kwargs.get("deployment")
        self._force_apply = kwargs.get("force_apply")
        self._show_output = kwargs.get("show_output")
        self._terraform_bin = kwargs.get("terraform_bin")

        self._plan_for = "destroy" if kwargs.get("destroy") else "apply"
        (self._tf_version_major, self._tf_version_minor) = kwargs.get(
            "tf_version", (None, None)
        )
        if self._tf_version_major is None or self._tf_version_minor is None:
            (
                self._tf_version_major,
                self._tf_version_minor,
            ) = self.get_terraform_version(self._terraform_bin)
        super(TerraformCommand, self).__init__(rootc, plan_for=self._plan_for, **kwargs)

    @property
    def plan_for(self):
        return self._plan_for

    def prep_modules(self):
        """Puts the modules sub directories into place."""
        mod_source = f"{self._repository_path}/terraform-modules".replace("//", "/")
        mod_destination = f"{self._temp_dir}/terraform-modules".replace("//", "/")
        click.secho(
            f"copying modules from {mod_source} to {mod_destination}", fg="yellow"
        )
        shutil.copytree(
            mod_source,
            mod_destination,
            symlinks=True,
            ignore=shutil.ignore_patterns("test", ".terraform", "terraform.tfstate*"),
        )

    def exec(self):
        for definition in self.definitions.limited():
            execute = False
            # copy definition files / templates etc.
            click.secho(f"preparing definition: {definition.tag}", fg="green")
            definition.prep(
                self._backend,
            )
            # run terraform init
            try:
                self._run(definition, "init", debug=self._show_output)
            except TerraformError:
                click.secho("error running terraform init", fg="red")
                raise SystemExit(1)

            click.secho(f"planning definition: {definition.tag}", fg="green")

            # run terraform plan
            try:
                self._run(
                    definition,
                    "plan",
                    debug=self._show_output,
                    plan_action=self._plan_for,
                )
            except PlanChange:
                execute = True
            except TerraformError:
                click.secho(
                    f"error planning terraform definition: {definition.tag}!",
                    fg="red",
                )
                raise SystemExit(2)

            if self._force_apply:
                execute = True

            if execute and self._tf_apply:
                if self._force_apply:
                    click.secho(
                        f"force apply for {definition.tag}, applying",
                        fg="yellow",
                    )
                else:
                    click.secho(
                        f"plan changes for {definition.tag}, applying",
                        fg="yellow",
                    )
            elif execute and self._destroy:
                click.secho(
                    f"plan changes for {definition.tag}, destroying",
                    fg="yellow",
                )
            elif not execute:
                click.secho(f"no plan changes for {definition.tag}", fg="yellow")
                continue

            try:
                self._run(
                    definition,
                    self._plan_for,
                    debug=self._show_output,
                )
            except TerraformError:
                click.secho(
                    f"error with terraform {self._plan_for} on definition"
                    f" {definition.tag}, exiting",
                    fg="red",
                )
                raise SystemExit(2)
            else:
                click.secho(
                    f"terraform {self._plan_for} complete for {definition.tag}",
                    fg="green",
                )

    def _run(self, definition, command, debug=False, plan_action="init"):
        """Run terraform."""
        if self._tf_version_major == 12:
            params = {
                "init": f"-input=false -no-color -plugin-dir={self._temp_dir}/terraform-plugins",
                "plan": "-input=false -detailed-exitcode -no-color",
                "apply": "-input=false -no-color -auto-approve",
                "destroy": "-input=false -no-color -force",
            }
        else:
            params = {
                "init": "-input=false -no-color",
                "plan": "-input=false -detailed-exitcode -no-color",
                "apply": "-input=false -no-color -auto-approve",
                "destroy": "-input=false -no-color -force",
            }

        if plan_action == "destroy":
            params["plan"] += " -destroy"

        env = os.environ.copy()
        for auth in self._authenticators:
            env.update(auth.env())

        env["TF_PLUGIN_CACHE_DIR"] = f"{self._temp_dir}/terraform-plugins"

        working_dir = f"{self._temp_dir}/definitions/{definition.tag}"
        command_params = params.get(command)
        if not command_params:
            raise ValueError(
                f"invalid command passed to terraform, {command} has no defined params!"
            )

        # only execute hooks for plan/apply/destroy
        try:
            if TerraformCommand.check_hooks(
                "pre", working_dir, command
            ) and command in ["apply", "destroy", "plan"]:
                # pre exec hooks
                # want to pass remotes
                # want to pass tf_vars
                click.secho(
                    f"found pre-{command} hook script for definition {definition.tag},"
                    " executing ",
                    fg="yellow",
                )
                TerraformCommand.hook_exec(
                    "pre",
                    command,
                    working_dir,
                    env,
                    self._terraform_bin,
                    debug=debug,
                    b64_encode=self._b64_encode,
                )
        except HookError as e:
            click.secho(
                f"hook execution error on definition {definition.tag}: {e}",
                fg="red",
            )
            raise SystemExit(2)

        click.secho(
            f"cmd: {self._terraform_bin} {command} {command_params}", fg="yellow"
        )
        (exit_code, stdout, stderr) = TerraformCommand.pipe_exec(
            f"{self._terraform_bin} {command} {command_params}",
            cwd=working_dir,
            env=env,
        )
        if debug:
            click.secho(f"exit code: {exit_code}", fg="blue")
            for line in stdout.decode().splitlines():
                click.secho(f"stdout: {line}", fg="blue")
            for line in stderr.decode().splitlines():
                click.secho(f"stderr: {line}", fg="red")

        # special handling of the exit codes for "plan" operations
        if command == "plan":
            if exit_code == 0:
                return True
            if exit_code == 1:
                raise TerraformError
            if exit_code == 2:
                raise PlanChange

        if exit_code:
            raise TerraformError

        # only execute hooks for plan/destroy
        try:
            if TerraformCommand.check_hooks(
                "post", working_dir, command
            ) and command in ["apply", "destroy", "plan"]:
                click.secho(
                    f"found post-{command} hook script for definition {definition.tag},"
                    " executing ",
                    fg="yellow",
                )
                TerraformCommand.hook_exec(
                    "post",
                    command,
                    working_dir,
                    env,
                    self._terraform_bin,
                    debug=debug,
                    b64_encode=self._b64_encode,
                )
        except HookError as e:
            click.secho(
                f"hook execution error on definition {definition.tag}: {e}", fg="red"
            )
            raise SystemExit(2)
        return True

    @staticmethod
    def hook_exec(
        phase,
        command,
        working_dir,
        env,
        terraform_path,
        debug=False,
        b64_encode=False,
    ):
        """
        hook_exec executes a hook script.

        Before execution it sets up the environment to make all terraform and remote
        state variables available to the hook via environment vars
        """

        key_replace_items = {
            " ": "",
            '"': "",
            "-": "_",
            ".": "_",
        }
        val_replace_items = {
            " ": "",
            '"': "",
            "\n": "",
        }
        local_env = env.copy()
        hook_dir = f"{working_dir}/hooks"
        hook_script = None

        for f in os.listdir(hook_dir):
            # this file format is specifically structured by the prep_def function
            if os.path.splitext(f)[0] == f"{phase}_{command}":
                hook_script = f"{hook_dir}/{f}"
        # this should never have been called if the hook script didn't exist...
        if hook_script is None:
            raise HookError(f"hook script missing from {hook_dir}")

        # populate environment with terraform remotes
        if os.path.isfile(f"{working_dir}/worker-locals.tf"):
            # I'm sorry. :-)
            r = re.compile(
                r"\s*(?P<item>\w+)\s*\=.+data\.terraform_remote_state\.(?P<state>\w+)\.outputs\.(?P<state_item>\w+)\s*"
            )

            with open(f"{working_dir}/worker-locals.tf") as f:
                for line in f:
                    m = r.match(line)
                    if m:
                        item = m.group("item")
                        state = m.group("state")
                        state_item = m.group("state_item")
                    else:
                        continue

                    state_value = TerraformCommand.get_state_item(
                        working_dir, env, terraform_path, state, state_item
                    )

                    if state_value is not None:
                        if b64_encode:
                            state_value = base64.b64encode(
                                state_value.encode("utf-8")
                            ).decode()
                        local_env[f"TF_REMOTE_{state}_{item}".upper()] = state_value

        # populate environment with terraform variables
        if os.path.isfile(f"{working_dir}/worker.auto.tfvars"):
            with open(f"{working_dir}/worker.auto.tfvars") as f:
                for line in f:
                    tf_var = line.split("=")

                    # strip bad names out for env var settings
                    for k, v in key_replace_items.items():
                        tf_var[0] = tf_var[0].replace(k, v)

                    for k, v in val_replace_items.items():
                        tf_var[1] = tf_var[1].replace(k, v)

                    if b64_encode:
                        tf_var[1] = base64.b64encode(tf_var[1].encode("utf-8")).decode()

                    local_env[f"TF_VAR_{tf_var[0].upper()}"] = tf_var[1]
        else:
            click.secho(
                f"{working_dir}/worker.auto.tfvars not found!",
                fg="red",
            )

        # execute the hook
        (exit_code, stdout, stderr) = TerraformCommand.pipe_exec(
            f"{hook_script} {phase} {command}",
            cwd=hook_dir,
            env=local_env,
        )

        # handle output from hook_script
        if debug:
            click.secho(f"exit code: {exit_code}", fg="blue")
            for line in stdout.decode().splitlines():
                click.secho(f"stdout: {line}", fg="blue")
            for line in stderr.decode().splitlines():
                click.secho(f"stderr: {line}", fg="red")

        if exit_code != 0:
            raise HookError("hook script {}")

    @staticmethod
    def pipe_exec(args, stdin=None, cwd=None, env=None):
        """
        A function to accept a list of commands and pipe them together.

        Takes optional stdin to give to the first item in the pipe chain.
        """
        count = 0
        commands = []
        if env is None:
            env = os.environ.copy()

        if not isinstance(args, list):
            args = [args]

        for i in args:
            if count == 0:
                if stdin is None:
                    commands.append(
                        subprocess.Popen(
                            shlex.split(i),
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            cwd=cwd,
                            env=env,
                        )
                    )
                else:
                    commands.append(
                        subprocess.Popen(
                            shlex.split(i),
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            stdin=subprocess.PIPE,
                            cwd=cwd,
                            env=env,
                        )
                    )
            else:
                commands.append(
                    subprocess.Popen(
                        shlex.split(i),
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        stdin=commands[count - 1].stdout,
                        cwd=cwd,
                        env=env,
                    )
                )
            count = count + 1

        if stdin is not None:
            stdin_bytes = stdin.encode()
            if len(commands) > 1:
                commands[0].communicate(input=stdin_bytes)
                stdout, stderr = commands[-1].communicate()
                commands[-1].wait()
                returncode = commands[-1].returncode
            else:
                stdout, stderr = commands[0].communicate(input=stdin_bytes)
                commands[0].wait()
                returncode = commands[0].returncode
        else:
            stdout, stderr = commands[-1].communicate()
            commands[-1].wait()
            returncode = commands[-1].returncode

        return (returncode, stdout, stderr)

    @staticmethod
    def check_hooks(phase, working_dir, command):
        """
        check_hooks determines if a hook exists for a given operation/definition
        """
        hook_dir = f"{working_dir}/hooks"
        if not os.path.isdir(hook_dir):
            # there is no hooks dir
            return False
        for f in os.listdir(hook_dir):
            if os.path.splitext(f)[0] == f"{phase}_{command}":
                if os.access(f"{hook_dir}/{f}", os.X_OK):
                    return True
                else:
                    raise HookError(f"{hook_dir}/{f} exists, but is not executable!")
        return False

    @staticmethod
    def get_state_item(working_dir, env, terraform_bin, state, item):
        """
        get_state_item returns json encoded output from a terraform remote state
        """
        base_dir, _ = os.path.split(working_dir)
        try:
            (exit_code, stdout, stderr) = TerraformCommand.pipe_exec(
                f"{terraform_bin} output -json -no-color {item}",
                cwd=f"{base_dir}/{state}",
                env=env,
            )
        except FileNotFoundError:
            # the remote state is not setup, likely do to use of --limit
            # this is acceptable, and is the responsibility of the hook
            # to ensure it has all values needed for safe execution
            return None

        if exit_code != 0:
            raise HookError(
                f"Error reading remote state item {state}.{item}, details: {stderr}"
            )

        if stdout is None:
            raise HookError(
                f"Remote state item {state}.{item} is empty; This is completely"
                " unexpected, failing..."
            )
        json_output = json.loads(stdout)
        return json.dumps(json_output, indent=None, separators=(",", ":"))

    @staticmethod
    def get_terraform_version(terraform_bin):
        (return_code, stdout, stderr) = TerraformCommand.pipe_exec(
            f"{terraform_bin} version"
        )
        if return_code != 0:
            click.secho(f"unable to get terraform version\n{stderr}", fg="red")
            raise SystemExit(1)
        version = stdout.decode("UTF-8").split("\n")[0]
        version_search = re.search(r".* v\d+\.(\d+)\.(\d+)", version)
        if version_search:
            click.secho(
                f"Terraform Version Result: {version}, using major:{version_search.group(1)}, minor:{version_search.group(2)}",
                fg="yellow",
            )
            return (int(version_search.group(1)), int(version_search.group(2)))
        else:
            click.secho(f"unable to get terraform version\n{stderr}", fg="red")
            raise SystemExit(1)



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

_CWD = os.getcwd()

DEFAULT_BACKEND_BUCKET = "tfworker-terraform-states"
DEFAULT_BACKEND_PREFIX = "terraform/state/{deployment}"
DEFAULT_CONFIG = f"{_CWD}/worker.yaml"
DEFAULT_REPOSITORY_PATH = _CWD
DEFAULT_AWS_REGION = "us-west-2"
DEFAULT_GCP_REGION = "us-west2b"
DEFAULT_BACKEND_REGION = "us-west-2"
DEFAULT_TERRFORM = "/usr/local/bin/terraform"

RESERVED_FILES = ["terraform.tf", "worker-locals.tf"]



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import pathlib
import platform
import re
import shutil
import tempfile
from collections import OrderedDict

import click
import jinja2
import yaml
from jinja2.runtime import StrictUndefined


class RootCommand:
    def __init__(self, args=None, clean=True):
        """Setup state with args that are passed."""
        self.clean = clean
        self.temp_dir = tempfile.mkdtemp()
        self.args = self.StateArgs()

        # Config accessors
        self.tf = None
        self._pullup_keys()

        if args is not None:
            self.add_args(args)

        if args.get("config_file") and args.get("backend"):
            click.secho(f"loading config file {args.get('config_file')}", fg="green")
            self.load_config(args.get("config_file"))

    def __del__(self):
        """Cleanup the temporary directory after execution."""
        if self.clean:
            try:
                shutil.rmtree(self.temp_dir)
            except Exception:
                pass

    def add_args(self, args):
        """Add a dictionary of args."""
        for k, v in args.items():
            self.add_arg(k, v)

    def add_arg(self, k, v):
        """Add an argument to the state args."""
        setattr(self.args, k, v)
        return None

    def load_config(self, config_file):
        try:
            template_reader = io.StringIO()
            jinja_env = jinja2.Environment(
                undefined=StrictUndefined,
                loader=jinja2.FileSystemLoader(pathlib.Path(config_file).parents[0]),
            )
            template_config = jinja_env.get_template(pathlib.Path(config_file).name)
            # maybe get_env should be optional?
            template_config.stream(
                **self.args.template_items(return_as_dict=True, get_env=True)
            ).dump(template_reader)
            self.config = ordered_config_load(template_reader.getvalue(), self.args)

            # A little arbitrary, but decorate the top two levels
            # directly on self object
            self.tf = self.config.get("terraform", OrderedDict())
            self._pullup_keys()
        except jinja2.exceptions.TemplateNotFound as e:
            path = pathlib.Path(config_file).parents[0]
            click.secho(f"can not read template file: {path}/{e}", fg="red")
            raise SystemExit(1)
        except jinja2.exceptions.UndefinedError as e:
            click.secho(
                f"configuration file contains invalid template substitutions: {e}",
                fg="red",
            )
            raise SystemExit(1)

    def _pullup_keys(self):
        """_pullup_keys is a utility function to place keys from the loaded config file
        directly on the RootCommand instance."""
        for k in [
            "definitions",
            "plugins",
            "providers",
            "remote_vars",
            "template_vars",
            "terraform_vars",
        ]:
            if self.tf:
                setattr(self, f"{k}_odict", self.tf.get(k, OrderedDict()))
            else:
                setattr(self, f"{k}_odict", None)

    class StateArgs:
        """A class to hold arguments in the state for easier access."""

        def __iter__(self):
            return iter(self.__dict__)

        def __getitem__(self, name):
            return self.__dict__[name]

        def keys(self):
            return self.__dict__.keys()

        def items(self):
            return self.__dict__.items()

        def values(self):
            return self.__dict__.values()

        def template_items(self, return_as_dict=False, get_env=False):
            rvals = {}
            for k, v in self.__dict__.items():
                if k == "config_var":
                    try:
                        rvals["var"] = get_config_var_dict(v)
                    except ValueError as e:
                        click.secho(
                            f'Invalid config-var specified: "{e}" must be in format key=value',
                            fg="red",
                        )
                        raise SystemExit(1)
                else:
                    rvals[k] = v
            if get_env is True:
                rvals["env"] = dict()
                for k, v in os.environ.items():
                    rvals["env"][k] = v
            if return_as_dict:
                return rvals
            return rvals.items()


def get_config_var_dict(config_vars):
    """
    get_config_var_dict returns a dictionary of of key=value for each item
    provided as a command line substitution
    """
    return_vars = dict()
    for cv in config_vars:
        try:
            k, v = tuple(cv.split("="))
            return_vars[k] = v
        except ValueError:
            raise ValueError(cv)
    return return_vars


def ordered_config_load(
    stream, args, Loader=yaml.SafeLoader, object_pairs_hook=OrderedDict
):
    """
    Load a yaml config, and replace templated items.

    Derived from:
    https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts
    """

    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        for item in node.value:
            if isinstance(item, tuple):
                for oneitem in item:
                    if isinstance(oneitem, yaml.ScalarNode):
                        oneitem.value = replace_vars(oneitem.value, args)

        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping
    )
    return yaml.load(stream, OrderedLoader)


def replace_vars(var, args):
    """Replace variables with template values."""
    var_pattern = r"//\s*(\S*)\s*//"
    match = re.match(var_pattern, var, flags=0)
    if not match:
        return var
    try:
        var = getattr(args, match.group(1).replace("-", "_"))
        click.secho(
            f"DEPRECATION WARNING: the use of //{match.group(1)}// substituion is deprecated and will be removed in a future version, use Jinja2 template style: {{{{ {match.group(1)} }}}}",
            fg="yellow",
        )
    except AttributeError:
        raise (ValueError(f"substitution not found for {var}"))
    return var


def get_platform():
    """
    get_platform will return a formatted operating system / architecture
    tuple that is consistent with common distribution creation tools
    """

    # strip off "2" which only appears on old linux kernels
    opsys = platform.system().rstrip("2").lower()

    # make sure machine uses consistent format
    machine = platform.machine()
    if machine == "x86_64":
        machine = "amd64"

    return (opsys, machine)



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class BaseBackend:
    tag = "base"

    def hcl(self, name):
        raise NotImplementedError

    def data_hcl(self, exclude):
        raise NotImplementedError


class Backends:
    s3 = "s3"
    gcs = "gcs"



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

from .aws import AWSAuthenticator  # noqa
from .base import UnknownAuthenticator  # noqa
from .google import GoogleAuthenticator  # noqa

ALL = [AWSAuthenticator, GoogleAuthenticator]


class AuthenticatorsCollection(collections.abc.Mapping):
    def __init__(self, state_args, **kwargs):
        self._authenticators = dict(
            [(auth.tag, auth(state_args, **kwargs)) for auth in ALL]
        )

    def __len__(self):
        return len(self._authenticators)

    def __getitem__(self, value):
        if type(value) == int:
            return self._authenticators[list(self._authenticators.keys())[value]]
        return self._authenticators[value]

    def __iter__(self):
        return iter(self._authenticators.values())

    def get(self, value):
        try:
            return self[value]
        except Exception:
            pass
        return None



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import copy
import os
import shutil
import sys
from pathlib import Path

import click
import jinja2
from tfworker import constants as const


class ReservedFileError(Exception):
    pass


class Definition:
    def __init__(
        self,
        definition,
        body,
        deployment,
        global_remote_vars,
        global_template_vars,
        global_terraform_vars,
        providers,
        repository_path,
        temp_dir,
        limited=False,
    ):
        self.tag = definition
        self._body = body
        self._path = body.get("path")
        self._remote_vars = self.make_vars(
            body.get("remote_vars", collections.OrderedDict()), global_remote_vars
        )
        self._template_vars = self.make_vars(
            body.get("template_vars", collections.OrderedDict()), global_template_vars
        )
        self._terraform_vars = self.make_vars(
            body.get("terraform_vars", collections.OrderedDict()), global_terraform_vars
        )

        self._deployment = deployment
        self._repository_path = repository_path
        self._providers = providers
        self._temp_dir = temp_dir
        self._limited = limited

    @property
    def body(self):
        return self._body

    @property
    def limited(self):
        return self._limited

    @property
    def path(self):
        return self._path

    def prep(self, backend):
        """ prepare the definitions for running """
        repo = Path(f"{self._repository_path}/{self.path}".replace("//", "/"))
        target = Path(f"{self._temp_dir}/definitions/{self.tag}".replace("//", "/"))
        target.mkdir(parents=True, exist_ok=True)

        if not repo.exists():
            click.secho(
                f"Error preparing definition {self.tag}, path {repo.resolve()} does not"
                " exist",
                fg="red",
            )
            sys.exit(1)

        # Prepare variables
        self._template_vars["deployment"] = self._deployment
        self._terraform_vars["deployment"] = self._deployment

        # Put terraform files in place
        for tf in repo.glob("*.tf"):
            if tf.name in const.RESERVED_FILES:
                raise ReservedFileError(f"{tf} is not allowed")
            shutil.copy(str(tf), str(target))

        for tf in repo.glob("*.tfvars"):
            shutil.copy(str(tf), str(target))

        if os.path.isdir(f"{repo}/templates".replace("//", "/")):
            shutil.copytree(f"{repo}/templates", f"{target}/templates")

        if os.path.isdir(f"{repo}/policies".replace("//", "/")):
            shutil.copytree(f"{repo}/policies", f"{target}/policies")

        if os.path.isdir(f"{repo}/scripts".replace("//", "/")):
            shutil.copytree(f"{repo}/scripts", f"{target}/scripts")

        if os.path.isdir(f"{repo}/hooks".replace("//", "/")):
            shutil.copytree(f"{repo}/hooks", f"{target}/hooks")

        if os.path.isdir(f"{repo}/repos".replace("//", "/")):
            shutil.copytree(f"{repo}/repos", f"{target}/repos")

        # Render jinja templates and put in place
        env = jinja2.Environment(loader=jinja2.FileSystemLoader)

        for j2 in repo.glob("*.j2"):
            contents = env.get_template(str(j2)).render(**self._template_vars)
            with open(f"{target}/{j2}", "w+") as j2_file:
                j2_file.write(contents)

        # Create local vars from remote data sources
        if len(list(self._remote_vars.keys())) > 0:
            with open(f"{target}/worker-locals.tf", "w+") as tflocals:
                tflocals.write("locals {\n")
                for k, v in self._remote_vars.items():
                    tflocals.write(f"  {k} = data.terraform_remote_state.{v}\n")
                tflocals.write("}\n\n")

        with open(f"{target}/terraform.tf", "w+") as tffile:
            tffile.write(f"{self._providers.hcl()}\n\n")
            tffile.write(f"{backend.hcl(self.tag)}\n\n")
            tffile.write(backend.data_hcl(self.tag))

        # Create the variable definitions
        with open(f"{target}/worker.auto.tfvars", "w+") as varfile:
            for k, v in self._terraform_vars.items():
                if isinstance(v, list):
                    varstring = f'[{", ".join(map(Definition.quote_str, v))}]'
                    varfile.write(f"{k} = {varstring}\n")
                else:
                    varfile.write(f'{k} = "{v}"\n')

    @staticmethod
    def quote_str(some_string):
        """Put literal quotes around a string."""
        return f'"{some_string}"'

    def make_vars(self, local_vars, global_vars):
        """Make a variables dictionary based on default vars, as well as specific vars for an item."""
        global_vars = global_vars or collections.OrderedDict()
        item_vars = copy.deepcopy(global_vars)
        for k, v in local_vars.items():
            # terraform expects variables in a specific type, so need to convert bools to a lower case true/false
            matched_type = False
            if v is True:
                item_vars[k] = "true"
                matched_type = True
            if v is False:
                item_vars[k] = "false"
                matched_type = True
            if not matched_type:
                item_vars[k] = v

        return item_vars


class DefinitionsCollection(collections.abc.Mapping):
    def __init__(
        self,
        definitions,
        deployment,
        limit,
        plan_for,
        providers,
        repository_path,
        rootc,
        temp_dir,
    ):
        self._body = definitions
        self._plan_for = plan_for
        self._definitions = collections.OrderedDict()
        for definition, body in definitions.items():
            self._definitions[definition] = Definition(
                definition,
                body,
                deployment,
                rootc.remote_vars_odict,
                rootc.template_vars_odict,
                rootc.terraform_vars_odict,
                providers,
                repository_path,
                temp_dir,
                True if limit and definition in limit else False,
            )

    def __len__(self):
        return len(self._definitions)

    def __getitem__(self, value):
        if type(value) == int:
            return self._definitions[list(self._definitions.keys())[value]]
        return self._definitions[value]

    def __iter__(self):
        return self.iter(honor_destroy=True)

    def iter(self, honor_destroy=False):
        if honor_destroy:
            if self._plan_for == "destroy":
                return iter(reversed(list(self._definitions.values())))
        return iter(self._definitions.values())

    def limited(self):
        # handle the case where nothing is filtered
        if len(list(filter(lambda d: d.limited, self.iter(honor_destroy=True)))) == 0:
            # the run is not limited to anything, so return everything
            return self.iter(honor_destroy=True)
        else:
            return iter(filter(lambda d: d.limited, self.iter(honor_destroy=True)))

    @property
    def body(self):
        return self._body



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import BaseBackend


class GCSBackend(BaseBackend):
    tag = "gcs"
    auth_tag = "google"

    def __init__(self, authenticators, definitions, deployment=None):
        self._authenticator = authenticators[self.auth_tag]
        self._definitions = definitions
        if deployment:
            self._deployment = deployment

    def hcl(self, name):
        state_config = []
        state_config.append("terraform {")
        state_config.append('  backend "gcs" {')
        state_config.append(f'    bucket = "{self._authenticator.bucket}"')
        state_config.append(f'    prefix = "{self._authenticator.prefix}/{name}"')
        if self._authenticator.creds_path:
            state_config.append(f'    credentials = "{self._authenticator.creds_path}"')
        state_config.append("  }")
        state_config.append("}")
        return "\n".join(state_config)

    def data_hcl(self, exclude):
        remote_data_config = []
        # Call the iter method for explicit control of iteration order
        for definition in self._definitions.iter():
            if definition.tag == exclude:
                break
            remote_data_config.append(
                f'data "terraform_remote_state" "{definition.tag}" {{'
            )
            remote_data_config.append('  backend = "gcs"')
            remote_data_config.append("  config = {")
            remote_data_config.append(f'    bucket = "{self._authenticator.bucket}"')
            remote_data_config.append(
                f'    prefix = "{self._authenticator.prefix}/{definition.tag}"'
            )
            if self._authenticator.creds_path:
                remote_data_config.append(
                    f'    credentials = "{self._authenticator.creds_path}"'
                )
            remote_data_config.append("  }")
            remote_data_config.append("}")
        return "\n".join(remote_data_config)



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class BaseProvider:
    tag = None

    def __init__(self, body):
        self.vars = body.get("vars", {})
        self.version = self.vars.get("version")

    def hcl(self):
        result = []
        provider_vars = {}
        try:
            for k, v in self.vars.items():
                provider_vars[k] = v
        except (KeyError, TypeError):
            """No provider vars were set."""
            pass

        result.append(f'provider "{self.tag}" {{')
        for k, v in provider_vars.items():
            if v and '"' not in v:
                result.append(f'  {k} = "{v}"')
            else:
                result.append(f"  {k} = {v}")
        result.append("}")
        return "\n".join(result)

    def clean(self, deployment, limit, config):
        """Nothing to do here so far"""
        pass


class UnknownProvider(Exception):
    def __init__(self, provider):
        super().__init__(f"{provider} is not a known value.")


class BackendError(Exception):
    pass


def validate_backend_empty(state):
    """
    validate_backend_empty ensures that the provided state file
    is empty
    """

    try:
        if len(state["resources"]) > 0:
            return False
        else:
            return True
    except KeyError:
        raise BackendError("resources key does not exist in state!")


def validate_backend_region(state):
    """
    validate_backend_region validates that a statefile
    was previously used in the region the current
    deployment is being created for
    """



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class BaseAuthenticator:
    tag = "base"

    def __init__(self, state_args, **kwargs):
        self._args = state_args
        self.clean = kwargs.get("clean")

    def _resolve_arg(self, name):
        return getattr(self._args, name) if hasattr(self._args, name) else None

    def env(self):
        return {}


class UnknownAuthenticator(Exception):
    def __init__(self, provider):
        super().__init__(f"{provider} is not a known value.")



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="terraform-worker",
    version="0.7.2",
    packages=find_packages(exclude=["tests*"]),
    author="Richard Maynard",
    author_email="richard.maynard@gmail.com",
    description="An orchestration tool for Terraform",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ephur/terraform-worker",
    include_package_data=True,
    install_requires=[
        "boto3",
        "hvac",
        "pyhcl",
        "tenacity",
        "cryptography",
        "click",
        "Jinja2",
        "PyYAML",
        "requests",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Topic :: System :: Systems Administration",
    ],
    entry_points={"console_scripts": ["worker=tfworker.cli:cli"]},
    setup_requires=["flake8"],
    python_requires=">=3.6",
)



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from contextlib import closing

import click

from .base import BackendError, BaseProvider, validate_backend_empty


class StateError(Exception):
    pass


class AWSProvider(BaseProvider):
    tag = "aws"

    def __init__(self, body, authenticators, *args, **kwargs):
        super(AWSProvider, self).__init__(body)

        self._authenticator = authenticators.get(self.tag)
        self.vars = body.get("vars", {})

    # Provider-specific methods
    def _clean_bucket_state(self, definition=None):
        """
        clean_state validates all of the terraform states are empty,
        and then removes the backend objects from S3

        optionally definition can be passed to limit the cleanup
        to a single definition
        """

        s3_paginator = self._authenticator.backend_session.client("s3").get_paginator(
            "list_objects_v2"
        )
        s3_client = self._authenticator.backend_session.client("s3")
        if definition is None:
            prefix = self._authenticator.prefix
        else:
            prefix = f"{self._authenticator.prefix}/{definition}"

        for s3_object in self.filter_keys(
            s3_paginator, self._authenticator.bucket, prefix
        ):
            backend_file = s3_client.get_object(
                Bucket=self._authenticator.bucket, Key=s3_object
            )
            body = backend_file["Body"]
            with closing(backend_file["Body"]):
                backend = json.load(body)

            if validate_backend_empty(backend):
                self.delete_with_versions(s3_object)
                click.secho(f"backend file removed: {s3_object}", fg="yellow")
            else:
                raise BackendError(f"backend at: {s3_object} is not empty!")

    def _clean_locking_state(self, deployment, definition=None):
        """
        clean_locking_state when called removes the dynamodb table
        that holds all of the state checksums and locking table
        entries
        """
        dynamo_client = self.state_session.resource("dynamodb")
        if definition is None:
            table = dynamo_client.Table(f"terraform-{deployment}")
            table.delete()
        else:
            # delete only the entry for a single state resource
            table = dynamo_client.Table(f"terraform-{deployment}")
            table.delete_item(
                Key={
                    "LockID": f"{self._authenticator.bucket}/{self._authenticator.prefix}/{definition}/terraform.tfstate-md5"
                }
            )

    def clean(self, deployment, limit):
        if limit:
            for limit_item in limit:
                click.secho(
                    "when using limit, dynamodb tables won't be completely dropped",
                    fg="yellow",
                )
                try:
                    # the bucket state deployment is part of the s3 prefix
                    self._clean_bucket_state(definition=limit_item)
                    # deployment name needs specified to determine the dynamo table
                    self._clean_locking_state(deployment, definition=limit_item)
                except StateError as e:
                    click.secho(f"error deleting state: {e}", fg="red")
                    raise SystemExit(1)
        else:
            try:
                self._clean_bucket_state()
            except StateError as e:
                click.secho(f"error deleting state: {e}")
                raise SystemExit(1)
            self._clean_locking_state(deployment)

    def delete_with_versions(self, key):
        """
        delete_with_versions should handle object deletions, and all references / versions of the object

        note: in initial testing this isn't required, but is inconsistent with how S3 delete markers, and the boto
        delete object call work there may be some selfurations that require extra handling.
        """
        s3_client = self.state_session.client("s3")
        s3_client.delete_object(Bucket=self.state_bucket, Key=key)

    @staticmethod
    def filter_keys(paginator, bucket_name, prefix="/", delimiter="/", start_after=""):
        """
        filter_keys returns just they keys that are needed
        primarily from: https://stackoverflow.com/questions/30249069/listing-contents-of-a-bucket-with-boto3
        """

        prefix = prefix[1:] if prefix.startswith(delimiter) else prefix
        start_after = (
            (start_after or prefix) if prefix.endswith(delimiter) else start_after
        )
        try:
            for page in paginator.paginate(
                Bucket=bucket_name, Prefix=prefix, StartAfter=start_after
            ):
                for content in page.get("Contents", ()):
                    yield content["Key"]
        except TypeError:
            pass



#!/usr/bin/env python
# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import struct
import sys

import click
from tfworker import constants as const
from tfworker.commands import CleanCommand, RootCommand, TerraformCommand
from tfworker.commands.root import get_platform


def validate_deployment(ctx, deployment, name):
    """Validate the deployment is an 8 char name."""
    if len(name) > 16:
        click.secho("deployment must be less than 16 characters", fg="red")
        raise SystemExit(2)
    return name


def validate_gcp_creds_path(ctx, path, value):
    if value:
        if not os.path.isabs(value):
            value = os.path.abspath(value)
        if os.path.isfile(value):
            return value
        click.secho(f"Could not resolve GCP credentials path: {value}", fg="red")
        raise SystemExit(3)


def validate_host():
    """Ensure that the script is being run on a supported platform."""
    supported_opsys = ["darwin", "linux"]
    supported_machine = ["amd64"]

    opsys, machine = get_platform()

    if opsys not in supported_opsys:
        click.secho(
            f"this application is currently not known to support {opsys}",
            fg="red",
        )
        raise SystemExit(2)

    if machine not in supported_machine:
        click.secho(
            f"this application is currently not known to support running on {machine} machines",
            fg="red",
        )

    if struct.calcsize("P") * 8 != 64:
        click.secho(
            "this application can only be run on 64 bit hosts, in 64 bit mode", fg="red"
        )
        raise SystemExit(2)

    return True


@click.group()
@click.option(
    "--aws-access-key-id",
    required=False,
    envvar="AWS_ACCESS_KEY_ID",
    help="AWS Access key",
    default=None,
)
@click.option(
    "--aws-secret-access-key",
    required=False,
    envvar="AWS_SECRET_ACCESS_KEY",
    help="AWS access key secret",
    default=None,
)
@click.option(
    "--aws-session-token",
    required=False,
    envvar="AWS_SESSION_TOKEN",
    help="AWS access key token",
    default=None,
)
@click.option(
    "--aws-role-arn",
    envvar="AWS_ROLE_ARN",
    help="If provided, credentials will be used to assume this role (complete ARN)",
    default=None,
    required=False,
)
@click.option(
    "--aws-region",
    envvar="AWS_DEFAULT_REGION",
    default=const.DEFAULT_AWS_REGION,
    help="AWS Region to build in",
)
@click.option(
    "--aws-profile",
    required=False,
    envvar="AWS_PROFILE",
    help="The AWS/Boto3 profile to use",
    default=None,
)
@click.option(
    "--gcp-region",
    envvar="GCP_REGION",
    default=const.DEFAULT_GCP_REGION,
    help="Region to build in",
)
@click.option(
    "--gcp-creds-path",
    required=False,
    envvar="GCP_CREDS_PATH",
    help=(
        "Relative path to the credentials JSON file for the service account to be used."
    ),
    default=None,
    callback=validate_gcp_creds_path,
)
@click.option(
    "--gcp-project",
    envvar="GCP_PROJECT",
    help="GCP project name to which work will be applied",
    default=None,
)
@click.option(
    "--config-file",
    default=const.DEFAULT_CONFIG,
    envvar="WORKER_CONFIG_FILE",
    required=True,
)
@click.option(
    "--repository-path",
    default=const.DEFAULT_REPOSITORY_PATH,
    envvar="WORKER_REPOSITORY_PATH",
    required=True,
    help="The path to the terraform module repository",
)
@click.option(
    "--backend",
    type=click.Choice(["s3", "gcs"]),
    help="State/locking provider. One of: s3, gcs",
)
@click.option(
    "--backend-bucket",
    default=const.DEFAULT_BACKEND_BUCKET,
    help="Region where terraform rootc/lock bucket exists",
)
@click.option(
    "--backend-prefix",
    default=const.DEFAULT_BACKEND_PREFIX,
    help="Region where terraform rootc/lock bucket exists",
)
@click.option(
    "--backend-region",
    default=const.DEFAULT_BACKEND_REGION,
    help="Region where terraform rootc/lock bucket exists",
)
@click.option(
    "--config-var",
    multiple=True,
    default=[],
    help='key=value to be supplied as jinja variables in config_file under "var" dictionary, can be specified multiple times',
)
@click.pass_context
def cli(context, **kwargs):
    """CLI for the worker utility."""
    validate_host()
    config_file = kwargs["config_file"]
    try:
        context.obj = RootCommand(args=kwargs)
    except FileNotFoundError:
        click.secho(f"configuration file {config_file} not found", fg="red", err=True)
        raise SystemExit(1)


@cli.command()
@click.option("--limit", help="limit operations to a single definition", multiple=True)
@click.argument("deployment", callback=validate_deployment)
@click.pass_obj
def clean(rootc, *args, **kwargs):  # noqa: E501
    """ clean up terraform state """
    # clean just items if limit supplied, or everything if no limit
    CleanCommand(rootc, *args, **kwargs).exec()
    sys.exit(0)


@cli.command()
@click.option(
    "--clean/--no-clean",
    default=True,
    help="clean up the temporary directory created by the worker after execution",
)
@click.option(
    "--apply/--no-apply",
    "tf_apply",
    default=False,
    help="apply the terraform configuration",
)
@click.option(
    "--force-apply/--no-force-apply",
    "force_apply",
    default=False,
    help="force apply without plan change",
)
@click.option(
    "--destroy/--no-destroy",
    default=False,
    help="destroy a deployment instead of create it",
)
@click.option(
    "--show-output/--no-show-output",
    default=False,
    help="shot output from terraform commands",
)
@click.option(
    "--terraform-bin",
    default=const.DEFAULT_TERRFORM,
    help="The complate location of the terraform binary",
)
@click.option(
    "--b64-encode-hook-values/--no--b64-encode-hook-values",
    "b64_encode",
    default=False,
    help=(
        "Terraform variables and outputs can be complex data structures, setting this"
        " open will base64 encode the values for use in hook scripts"
    ),
)
@click.option("--limit", help="limit operations to a single definition", multiple=True)
@click.argument("deployment", callback=validate_deployment)
@click.pass_obj
def terraform(rootc, *args, **kwargs):
    """ No do nothing """
    tfc = TerraformCommand(rootc, *args, **kwargs)

    click.secho(f"building deployment {kwargs.get('deployment')}", fg="green")
    click.secho(f"using temporary Directory: {tfc.temp_dir}", fg="yellow")

    # common setup required for all definitions
    click.secho("downloading plugins", fg="green")
    tfc.plugins.download()
    click.secho("preparing modules", fg="green")
    tfc.prep_modules()

    tfc.exec()
    sys.exit(0)


if __name__ == "__main__":
    cli()
